/*
 *  ulocks.h   :  Fast User Level Synchronization 
 *                architecture specific routines
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (c) International Business Machines Corp., 2001
 *
 * Author: Hubertus Franke <frankeh@us.ibm.com, frankeh@watson.ibm.com>
 *
 */

#ifndef __ULOCKS_ASM_H__
#define __ULOCKS_ASM_H__

/*********************************************************************
 * BASIC ATOMIC OPREATIONS
 *********************************************************************/

struct __xchg_dummy { unsigned long a[99]; };
#define __xg(x) ((struct __xchg_dummy *)(x))

/**
 * __u_atomic_inc(unsigned long *addr)
 *    atomically increment (*addr)
 */

static inline 
void u_atomic_inc(unsigned long *addr)
{
	__asm__ __volatile__ ("lock; incl %0\n"	: "=m"(*addr));
}

/**
 * __ulock_down(ulock_t *v)
 *    atomically decrement (*v) and returns 1 if "< 0" and 0 otherwise.
 */

static inline
int __ulock_down(ulock_t *u)
{
	unsigned char c;
        __asm__ __volatile__(
                "lock ; decl %0; setl %1"
                :"=m" (u->status), "=qm" (c)
                :"m" (u->status) : "memory");
        return (c);
}

/**
 * __ulock_up(ulock_t *v)
 *    atomically increment (*v) and return 1 if "<= 0"
 */

static inline
int __ulock_up(ulock_t *u)
{
        unsigned char rc;
 
        __asm__ __volatile__(
                "lock ; incl %0\n"
		"setle %1\n"
                :"=m" (u->status), "=qm" (rc)
                :"m" (u->status) : "memory");
        return (rc);
}


/**
 * __urwlock_read_down(ulockt_t *v)
 *    atomically increment (*v) and return 1 if "<= 0"
 */

static inline
int __urwlock_read_down(ulock_t *u)
{
        unsigned char rc;
        __asm__ __volatile__(
                "lock ; incl %0; setl %1"
                :"=m" (u->status), "=qm" (rc)
                :"m" (u->status) : "memory");
        return (rc);
}

/**
 * compare_and_swap()
 *    compare value under addr with <oldvalue> and if equal atomically swaps
 *    value with <newvalue>. <oldvalue> is returned by the routine.
 */

static inline
unsigned long __compare_and_swap(volatile void *addr,
				 unsigned long oldvalue,
				 unsigned long newvalue)
{
	unsigned long prev;
	__asm__ __volatile__("lock ; cmpxchgl %1,%2"
                                     : "=a"(prev)
                                     : "q"(newvalue), 
			               "m"(*__xg(addr)), "0"(oldvalue)
                                     : "memory");
	return prev;
}

#endif
